#include <iostream>
#include "Graph.h"
#include "StackGraph.h"
#include "StackGraphItem.h"
#include "NodeArray.h"
#include "NodeStack.h"

using namespace std;

Graph *graph = NULL;

NodeArray *n_array = NULL;
NodeStack *n_stack = new NodeStack();

int *neighbours = NULL;
bool *bestSolution = NULL;
int minEdges = 0;

bool isBipartit(bool * potent) {
  n_array->reset();
  n_stack->reset();
  
  int fresh = n_array->getFresh();
  int color = 1;
  
  while(fresh != NodeArray::NOT_FOUND) {
    n_array->setColor(fresh, color);
    n_stack->push(fresh);
    
    while(!n_stack->empty()) {
      int aux = n_stack->pop();
      graph->getNeighbours(aux, potent, neighbours);
      color = 3 - n_array->getColor(aux);
      int i = 0;
      
      while(neighbours[i] != -1) {
        if(n_array->getColor(neighbours[i]) == 0) {
          n_array->setColor(neighbours[i], color);
          n_stack->push(neighbours[i]);
        }
        else if(n_array->getColor(neighbours[i]) != color) {
          return false;
        }
        
        i++;
      }
    }
    
    fresh = n_array->getFresh();
  }

  return true;
}

int solve() {
  StackGraph stack;
  int edges = graph->getEdgesCount();
  bool *first = new bool[edges];
  
  for(int i = 0; i < edges; i++) {
    first[i] = true;
  }
  
  stack.push(first, edges);
  
  while(!stack.empty()) {
    StackGraphItem *item = stack.pop();
    edges = item->getEdges();
    
    if(edges > minEdges) {
      bool *test = item->getGraph();
      if(isBipartit(test)) {
        delete [] bestSolution;
        bestSolution = test;
        minEdges = edges;
      }
      else {
        delete[] test;
        int first = item->getLastChange() + 1;
        for(int i = first; i < graph->getEdgesCount(); i++) {
          bool *potent = item->getGraph();
          if(potent[i]) {
            potent[i] = false;
            stack.push(potent, edges - 1);
          }
          else delete [] potent;
        }
      }
    }
    
    delete item;
  }
}

int printBestSolution() {
  cout << "Pocet hran: " << minEdges << "\nHrany:\n";
  
  for(int i = 0; i < graph->getEdgesCount(); i++)
    cout << bestSolution[i];

  cout << endl << endl;
  graph->print(bestSolution);
}

int main(int argc, char *argv[]) {
  if(argc != 2)
    cout << "Chyba: Musite uvest jmeno souboru s grafem\n";
  
  graph = new Graph(argv[1]);
  n_array = new NodeArray(graph->getNodesCount());
  neighbours = new int[graph->getNodesCount()];
  bestSolution = new bool[graph->getEdgesCount()];
  
  for(int i = 0; i < graph->getEdgesCount(); i++)
    bestSolution[i] = false;
  
  time_t start, end;
  time(&start);
  solve();
  time(&end);
  printBestSolution();
  cout << "Doba vypoctu: " << difftime(end,start) << endl;

  return 0;
}

